@@ -1,16 +1,12 @@
 #!/usr/bin/env python
 
 """
-Talk to a T616/T610 via a serial connection and make it do various things.
-
-For Bluetooth under Linux, you will need to bind a serial device before using
-this software - eg. rfcomm bind /dev/rfcomm0 PP:QQ:RR:XX:YY:ZZ - and then
-specify that device (eg. /dev/rfcomm0) in the port parameter of the T616
-constructor function.
+Talk to a T616/T610 via a serial or Bluetooth socket connection and make it do
+various things.
 
 Copyright (c) 2003 Nelson Minar <nelson@monkey.org>
 http://www.nelson.monkey.org/~nelson/weblog/tech/phone/
-Copyright (c) 2005 Paul Boddie <paul@boddie.org.uk>
+Copyright (c) 2005, 2006 Paul Boddie <paul@boddie.org.uk>
 http://www.boddie.org.uk/paul
 
 This library is free software; you can redistribute it and/or
@@ -28,7 +24,7 @@ License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 """
 
-__version__ = "0.1"
+__version__ = "0.2.1"
 
 try:
     import serial # Requires pySerial http://pyserial.sourceforge.net/
@@ -37,14 +33,16 @@ except ImportError:
 import socket, select
 import time, re, sys
 
+# Contact parsing functionality.
+
 contactParseRE = re.compile(r"""
   ^\+CPBR:\s
   (?P<pos>\d+),
-  "(?P<num>[\dp*#+]+)",
+  "(?P<num>[\dA-Fp*#+]+)",
   (?P<zone>\d+),
-  "(?P<name>[^/]+)(/(?P<type>\S))?"$""", re.VERBOSE)
+  "(?P<name>[^"]+)"$""", re.VERBOSE)
 
-def contactParse(s):
+def contactParse(s, charset):
 
     """
     Parse a contact from the phone's format. For example:
@@ -54,15 +52,22 @@ def contactParse(s):
     Return a tuple of strings describing the contact in the following format:
 
     (position, number, zone, name, type)
+
+    Note that type may be None if no type details were found.
     """
 
     m = contactParseRE.match(s)
     if not m:
         raise CommandException("Could not parse contact string %s" % repr(s))
     d = m.groupdict()
-    return d["pos"], d["num"], d["zone"], d["name"], d["type"]
+    name_and_type = from_phone(d["name"], charset)
+    if len(name_and_type) > 1 and name_and_type[-2] == "/":
+        name, type = name_and_type[:-2], name_and_type[-1]
+    else:
+        name, type = name_and_type, None
+    return d["pos"], from_phone(d["num"], charset), d["zone"], name, type
 
-def contactUnparse(t):
+def contactUnparse(t, charset):
 
     """
     Return a string like the phone's format. For example:
@@ -70,11 +75,15 @@ def contactUnparse(t):
     4,"5551212",129,"Test/H"
     """
 
+    position, number, zone, name, type = t
+
     # Check the type.
-    if t[-1] is not None:
-        return '%s,"%s",%s,"%s/%s"' % t
+    if type is not None:
+        return '%s,"%s",%s,"%s"' % (position, to_phone(number, charset), zone, to_phone(name + "/" + type, charset))
     else:
-        return '%s,"%s",%s,"%s"' % t[:-1]
+        return '%s,"%s",%s,"%s"' % (position, to_phone(number, charset), zone, to_phone(name, charset))
+
+# Exception classes, available on the T616 class itself.
 
 class CommandException(Exception):
 
@@ -88,6 +97,8 @@ class InitException(Exception):
 
     pass
 
+# Communications helper classes and functions.
+
 class SocketWrapper:
 
     """
@@ -134,11 +145,64 @@ class SocketWrapper:
 
         self.socket.close()
 
+    def getTimeout(self):
+        return self.timeout
+
+    def setTimeout(self, timeout):
+        self.timeout = timeout
+
     # Faked support for the clearing operation.
 
     def inWaiting(self):
         return 0
 
+# Character set helper functions.
+# NOTE: Should possibly consider genuine Unicode codecs.
+# NOTE: Should also fix GSM alphabet support.
+
+def to_phone(s, charset):
+
+    "Encode the given Unicode object 's' using the given 'charset'."
+
+    if charset == "UCS2":
+        return to_ucs2(s)
+    elif charset is None or charset in ("GSM", "IRA"):
+        return s.encode("ascii")
+    else:
+        return s.encode(charset)
+
+def from_phone(s, charset):
+
+    "Return the given string 's', encoded with 'charset', as a Unicode object."
+
+    if charset == "UCS2":
+        return from_ucs2(s)
+    elif charset is None or charset in ("GSM", "IRA"):
+        return unicode(s, "ascii")
+    else:
+        return unicode(s, charset)
+
+def to_ucs2(s):
+
+    "Encode the given Unicode object 's' as a UCS-2 string."
+
+    l = []
+    for c in unicode(s):
+        l.append(("%04x" % ord(c)).upper())
+    return "".join(l)
+
+def from_ucs2(s):
+
+    "Return the given UCS-2 string 's' as a Unicode object."
+
+    l = []
+    for i in range(0, len(s), 4):
+        c = s[i:i+4]
+        l.append(unichr(int(c, 16)))
+    return "".join(l)
+
+# The core functionality.
+
 class T616:
 
     """
@@ -161,9 +225,9 @@ class T616:
       with Bluetooth sockets
 
     More info:
+    http://www.boddie.org.uk/python/t616hack.html
     http://www.nelson.monkey.org/~nelson/weblog/tech/phone/
     http://www.ericsson.hu/mobilinternet/products/sh888/888_r1d.pdf
-    http://www.boddie.org.uk/python/t616hack.html
     """
 
     # Message status constants.
@@ -180,7 +244,7 @@ class T616:
     CommandException = CommandException
     InitException = InitException
 
-    def __init__(self, port=None, baudrate=115200, bdaddr=None, channel=1,
+    def __init__(self, port=None, baudrate=None, bdaddr=None, channel=None,
         timeout=1.0, debug=0, clear=0):
 
         """
@@ -209,6 +273,10 @@ class T616:
         """
 
         self.debug = debug
+        if channel is None:
+            channel = 1
+        if baudrate is None:
+            baudrate = 115200
         
         # Open the socket, port or device.
         # Check for a Bluetooth address.
@@ -222,7 +290,7 @@ class T616:
                 self.socket.close()
                 raise
 
-        # Check for pySerial
+        # Check for pySerial.
 
         elif serial is not None and port is not None:
             self.socket = None
@@ -247,6 +315,11 @@ class T616:
 
         self._expect('AT', '\r\nOK\r\n')
 
+        # Finally, set the character set.
+
+        self.charset = None
+        self.charset = self.getCharacterSet()
+
     def _debug(self, s):
         if self.debug: sys.stderr.write(":: %s\n" % s)
         
@@ -292,13 +365,15 @@ class T616:
         if not result == expected:
             raise CommandException("Expected %s but got %s" % (repr(expected), repr(result)))
 
+    # Public methods.
+
     def getTime(self):
 
         "Ask the phone what time it thinks it is."
 
         self._sendCmd("AT+CCLK?")
         r = self._readAll()
-        return re.search(r'CCLK: "(.+)"', r).group(1)
+        return from_phone(re.search(r'CCLK: "(.+)"', r).group(1), self.charset)
 
     def getVersion(self):
 
@@ -314,6 +389,7 @@ class T616:
         "SM".
         """
 
+        type = to_phone(type, self.charset)
         self._expect('AT+CPBS="%s"' % type, '\r\nOK\r\n')
 
     selectPhoneBookStorage = selectMemory
@@ -335,14 +411,14 @@ class T616:
 
         self._sendCmd("AT+CPMS=?")
         cpmsInfo = self._readAll()
-        m = re.search(r"\+CPMS: \((.+)\),\((.+)\),\((.+)\)", cpmsInfo)
+        m = re.search(r'\+CPMS: \((.+)\),\((.+)\),\((.+)\)', cpmsInfo)
         if not m:
             raise CommandException("Couldn't parse CPMS info %s" % repr(cpmsInfo))
         results = []
         for group in m.groups():
             result = []
             for identifier in group.split(","):
-                result.append(identifier.strip('"'))
+                result.append(from_phone(identifier.strip('"'), self.charset))
             results.append(result)
 
         # Cache the result.
@@ -360,22 +436,24 @@ class T616:
 
         self._sendCmd("AT+CPMS?")
         cpmsInfo = self._readAll()
-        m = re.search(r'\+CPMS: "([A-Z]+)",(\d+),(\d+),"([A-Z]+)",(\d+),(\d+),"([A-Z]+)",(\d+),(\d+)', cpmsInfo)
+        m = re.search(r'\+CPMS: "([A-Z0-9]+)",(\d+),(\d+),"([A-Z0-9]+)",(\d+),(\d+),"([A-Z0-9]+)",(\d+),(\d+)', cpmsInfo)
         if not m:
             raise CommandException("Couldn't parse CPMS info %s" % repr(cpmsInfo))
         g = m.groups()
         results = []
         for i in range(0, len(g), 3):
-            result = g[i], int(g[i+1]), int(g[i+2])
+            result = from_phone(g[i], self.charset), int(g[i+1]), int(g[i+2])
             results.append(result)
         return tuple(results)
 
-    def selectMessageStorage(self, type="ME"):
+    def selectMessageStorage(self, type="ME", mem2=None, mem3=None):
 
         """
         Select the preferred message storage 'type', which can typically either
-        be "ME" or "SM", returning a tuple containing a number of 2-tuple
-        elements, each element having the following format:
+        be "ME" or "SM", along with the optional 'mem2' and 'mem3' values
+        (selecting the second and third message storage types), returning a
+        tuple containing a number of 2-tuple elements, each element having the
+        following format:
 
         used_spaces_integer, total_spaces_integer
 
@@ -383,7 +461,10 @@ class T616:
         descending preference.
         """
 
-        self._sendCmd('AT+CPMS="%s"' % type)
+        cmd = 'AT+CPMS="%s"' % to_phone(type, self.charset)
+        if mem2: cmd += ',"%s"' % to_phone(mem2, self.charset)
+        if mem3: cmd += ',"%s"' % to_phone(mem3, self.charset)
+        self._sendCmd(cmd)
         cpmsInfo = self._readAll()
         m = re.search(r"\+CPMS: (\d+),(\d+),(\d+),(\d+),(\d+),(\d+)", cpmsInfo)
         if not m:
@@ -421,7 +502,7 @@ class T616:
         r = []
         for c in contacts:
             try:
-                r.append(contactParse(c))
+                r.append(contactParse(c, self.charset))
             except Exception, e:
                 # NOTE: This should be refined to catch specific exceptions.
                 # Continue past parse errors.
@@ -434,7 +515,7 @@ class T616:
 
         "Write the contact tuple 'c' to the phone."
 
-        self._expect("AT+CPBW=%s" % contactUnparse(c), "\r\nOK\r\n")
+        self._expect("AT+CPBW=%s" % contactUnparse(c, self.charset), "\r\nOK\r\n")
 
     def getMessageFormats(self, cache=1):
 
@@ -489,7 +570,7 @@ class T616:
         cscsInfo = self._readAll()
 
         # Cache the result.
-        self.character_sets = self._parseList("CSCS", cscsInfo)
+        self.character_sets = [from_phone(s, self.charset) for s in self._parseList("CSCS", cscsInfo)]
         return self.character_sets
 
     def getCharacterSet(self):
@@ -498,7 +579,7 @@ class T616:
 
         self._sendCmd("AT+CSCS?")
         cscsInfo = self._readAll()
-        return self._parseGeneral("CSCS", cscsInfo)
+        return from_phone(self._parseGeneral("CSCS", cscsInfo), self.charset)
 
     def setCharacterSet(self, charset):
 
@@ -507,7 +588,8 @@ class T616:
         possibly obtained from getCharacterSets.
         """
 
-        self._expect('AT+CSCS="%s"' % charset, '\r\nOK\r\n')
+        self._expect('AT+CSCS="%s"' % to_phone(charset, self.charset), '\r\nOK\r\n')
+        self.charset = charset
 
     def getMessageStatusCodes(self):
 
@@ -522,7 +604,7 @@ class T616:
             min, max = self._parseRange("CMGL", cmglInfo, "-")
             return range(min, max + 1)
         else:
-            return self._parseList("CMGL", cmglInfo)
+            return [from_phone(code, self.charset) for code in self._parseList("CMGL", cmglInfo)]
 
     def getApplicationFunctions(self):
 
@@ -557,16 +639,28 @@ class T616:
         returned from this method.
         """
 
+        return self._readMessagesRaw(code)
+
+    def _checkMessageStatusCode(self, code):
+
         # Get an appropriate request code.
         if code is None:
             code = self.getMessageStatusCodes()[-1]
         elif code not in self.getMessageStatusCodes():
             raise CommandException("Code %s was apparently not supported" % code)
+        return code
+
+    def _readMessagesRaw(self, code):
+
+        "Read the raw messages using the given 'code'."
+
+        code = self._checkMessageStatusCode(code)
 
         # Send the code and interpret the response.
         if type(code) == type(0):
             self._sendCmd("AT+CMGL=%s" % code)
         else:
+            code = to_phone(code, self.charset)
             self._sendCmd('AT+CMGL="%s"' % code)
         response = self._readAll()
         return response
@@ -592,17 +686,25 @@ class T616:
         Messages are returned as tuples each containing the message details and
         the raw response string."""
 
-        messageLines = self.readMessagesRaw(code).split("\r\n")[1:-3]
+        messageLines = self._readMessagesRaw(code).split("\r\n")[1:-3]
         messages = []
         i = 0
         while i < len(messageLines):
             messageDetails = self._parseHeader("CMGL", messageLines[i])
+
+            # Convert strings in the message details to Unicode.
+
+            messageDetails = [messageDetails[0]] + [from_phone(field, self.charset) for field in messageDetails[1:]]
+
+            # Collect the body lines together and combine the details and body.
+
             messageBodyLines = []
             i += 1
             while i < len(messageLines) and not messageLines[i].startswith("+CMGL"):
                 messageBodyLines.append(messageLines[i])
                 i += 1
-            messages.append((messageDetails, "".join(messageBodyLines)))
+            messages.append((messageDetails, from_phone("".join(messageBodyLines), self.charset)))
+
         return messages
 
     def readMessageRaw(self, index):
@@ -627,12 +729,20 @@ class T616:
 
         messageLines = self.readMessageRaw(index).split("\r\n")[1:-3]
         messageDetails = self._parseHeader("CMGR", messageLines[0])
+
+        # Convert strings to Unicode for all fields.
+
+        messageDetails = [from_phone(field, self.charset) for field in messageDetails]
+
+        # Collect the body lines together and return the combined details and
+        # body.
+
         i = 1
         messageBodyLines = []
         while i < len(messageLines) and not messageLines[i].startswith("+CMGR"):
             messageBodyLines.append(messageLines[i])
             i += 1
-        return (messageDetails, "".join(messageBodyLines))
+        return (messageDetails, from_phone("".join(messageBodyLines), self.charset))
 
     def deleteMessage(self, index):
 
@@ -643,6 +753,49 @@ class T616:
 
         self._expect("AT+CMGD=%s" % index, "\r\nOK\r\n")
 
+    def writeMessage(self, message, sender="", timeout=5):
+
+        """
+        Write the given 'message' (a Unicode object), optionally specifying the
+        'sender' (a Unicode object containing an international telephone number
+        without any non-numeric characters, such as the leading "+" character).
+        The optional 'timeout' (in seconds, default 5 seconds) specifies the
+        length of time that should be used to wait for the storage of the
+        message.
+        """
+
+        # Switch character set and message format.
+
+        format = self.getMessageFormat()
+        charset = self.getCharacterSet()
+        self.setMessageFormat(self.TEXT)
+        self.setCharacterSet("UCS2")
+
+        try:
+            self.ser.write('AT+CMGW')
+            self.ser.write('="%s"\r' % to_ucs2(sender))
+            response = self._readAll()
+            if response != '\r\n> \x00':
+                raise CommandException("Could not understand response %s" % repr(response))
+            self.ser.write(to_ucs2(message) + chr(26))
+            self._debug("(Message sent using AT+CMGW...)")
+
+            # Wait longer than usual for the result.
+
+            old_timeout = self.ser.getTimeout()
+            self.ser.setTimeout(timeout * 1000)
+            response = self._readAll()
+            self.ser.setTimeout(old_timeout)
+            if response != '\r\nOK\r\n':
+                raise CommandException("Could not understand response %s" % repr(response))
+
+        finally:
+
+            # Reset the character set and message format.
+
+            self.setCharacterSet(charset)
+            self.setMessageFormat(format)
+
     # Internal parsing methods.
 
     def _parseGeneral(self, command, response):
